/*
 * wrapconnection.cxx
 *
 * OpenH323 Wrapper Library
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wrapconnection.cxx,v 1.11 2003/07/21 10:52:35 manousos Exp $
 *
 */

/************************************************************************/

#include <ptlib.h>
#include <h323.h>
#include <h323pdu.h>
#include <mediafmt.h>
#include <lid.h>

#define cplusplus
#include "wrapper.hxx"
#include "wrapendpoint.hxx"
#include "wrapconnection.hxx"
#include "asteriskaudio.hxx"
#include "version.h"

using namespace std;

// External stuff used inside the wrapper library
extern int		channelsOpen;
extern PMutex	*channelMutex;
extern BOOL		Con_noFastStart;
extern BOOL		Con_noH245Tunnelling;
extern BOOL		Con_noH245InSetup;
//extern BOOL		Con_noInBandDTMF;
extern WORD		Con_jitterMin;
extern WORD		Con_jitterMax;
extern BOOL		Con_noSilenceSuppression;
extern start_logchan_cb		on_start_logical_channel; 
extern clear_con_cb			on_connection_cleared;
extern alert_con_cb			on_connection_alert;
extern h323_exception_cb	on_h323_exception;
extern init_con_cb			on_connection_init;

/********************************************
 * Class WrapH323Connection
 ********************************************/

WrapH323Connection::WrapH323Connection(WrapH323EndPoint & ep,
				   unsigned callReference)
: H323Connection(ep, callReference)
{
	// For outgoing calls: 
	//     Remove all capabilities and add only the one requested
	//     Set the caller ID
	if (ep.MakingCall) {
		if (ep.MakeCallCapability != NULL) {
			localCapabilities.RemoveAll();
			localCapabilities.SetCapability(0, 0, ep.MakeCallCapability);
			WRAPTRACE(3, "Outgoing call with capability " << 
						(ep.MakeCallCapability)->GetFormatName());
		}
		if (strlen(ep.MakeCallCID) != 0) {
			PString  cid = PString(ep.MakeCallCID);
			SetLocalPartyName(cid);
			WRAPTRACE(3, "Caller ID on outgoing call " << cid);
			WRAPTRACE(3, "LocalPartyName " << localPartyName);
			WRAPTRACE(3, "DestExtraCallInfo " << destExtraCallInfo);
		}
	}

	WRAPTRACE(1, "WrapH323Connection created.");
	return;
}

WrapH323Connection::~WrapH323Connection()
{
	WRAPTRACE(1, "WrapH323Connection deleted.");
	return;
}

H323Connection::AnswerCallResponse
WrapH323Connection::OnAnswerCall(const PString & caller,
					const H323SignalPDU & /*setupPDU*/,
					H323SignalPDU & /*connectPDU*/)
{
	call_details_t	cdet;		// Call details structure is passed to the PBX app
	int				res;

	// All the incoming calls receive an ALERTING PDU.
	// The call can be answered later with "AnsweringCall()" function.
	WRAPTRACE(2, "User " << caller << " is calling us...");

	// Fill the structure
	cdet.call_reference = GetCallReference();
	cdet.call_token = (const char *)GetCallToken();
	cdet.call_source_alias = (const char *)sourceAliases;
	cdet.call_dest_alias = (const char *)destAliases;
	cdet.call_source_e164 = (const char *)sourceE164;
	cdet.call_dest_e164 = (const char *)destE164;
	// DEBUG START
	WRAPTRACE(3, "Call reference: " << cdet.call_reference);
	WRAPTRACE(3, "Call token: " << cdet.call_token);
	WRAPTRACE(3, "Call source alias: "<< cdet.call_source_alias<<"("<<strlen(cdet.call_source_alias)<<")");
	WRAPTRACE(3, "Call dest alias: "<< cdet.call_dest_alias<<"("<<strlen(cdet.call_dest_alias)<<")");
	WRAPTRACE(3, "Call source e164: "<< cdet.call_source_e164<<"("<<strlen(cdet.call_source_e164)<<")");
	WRAPTRACE(3, "Call dest e164: "<< cdet.call_dest_e164<<"("<<strlen(cdet.call_dest_e164)<<")");
	WRAPTRACE(3, "Remote Party number: "<<GetRemotePartyNumber());
	WRAPTRACE(3, "Remote Party name: "<<GetRemotePartyName());
	WRAPTRACE(3, "Remote Party address: "<<GetRemotePartyAddress());
	// DEBUG END

	// Invoke the application callback
	res = on_connection_init(cdet);
	if (res < 0) {
		WRAPTRACE(2, "Failed to initialize incoming H.323 call. Dropping it.");
		return H323Connection::AnswerCallDenied;
	}

	return H323Connection::AnswerCallPending;
}

BOOL 
WrapH323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
		const PString & username)
{
	call_details_t cdet;

	WRAPTRACE(2, "Ringing phone for \"" << username << "\" ...");

	// Invoke the PBX application registered callback
	cdet.call_reference = GetCallReference();
	cdet.call_token = (const char *)GetCallToken();
	on_connection_alert(cdet);

	return TRUE;
}

BOOL 
WrapH323Connection::OnReceivedSignalSetup(const H323SignalPDU & setupPDU)
{
	WRAPTRACE(2, "Received SETUP message...");

	/***XXX Too early to do it 
	if (((WrapH323EndPoint &)GetEndPoint()).GetBandwidthAvailable() == 0) {	// XXX Temporary, fix it
		cout << "Bandwidth limit violation !!!" << endl;
		return FALSE;
	}
	************/

	sourceAliases = setupPDU.GetSourceAliases();
	destAliases = setupPDU.GetDestinationAlias();
	sourceE164 = "";
	setupPDU.GetSourceE164(sourceE164);
	destE164 = "";
	setupPDU.GetDestinationE164(destE164);

	if (H323Connection::OnReceivedSignalSetup(setupPDU)==FALSE){
		WRAPTRACE(2, "Returning FALSE.");
		return FALSE;
	}
	return TRUE;
}

BOOL 
WrapH323Connection::OnSendSignalSetup(H323SignalPDU & setupPDU)
{
	WRAPTRACE(2, "Sending SETUP message...");

	/***XXX Too early to do it 
	if (((WrapH323EndPoint &)GetEndPoint()).GetBandwidthAvailable() == 0) {	// XXX Temporary, fix it
		cout << "Bandwidth limit violation !!!" << endl;
		return FALSE;
	}
	************/

	sourceAliases = setupPDU.GetSourceAliases();
	destAliases = setupPDU.GetDestinationAlias();
	sourceE164 = "";
	setupPDU.GetSourceE164(sourceE164);
	destE164 = "";
	setupPDU.GetDestinationE164(destE164);

	return H323Connection::OnSendSignalSetup(setupPDU);
}

void
WrapH323Connection::OnReceivedReleaseComplete(const H323SignalPDU & pdu)
{
	WRAPTRACE(2, "Received RELEASE COMPLETE message...");

	return H323Connection::OnReceivedReleaseComplete(pdu);
}

BOOL
WrapH323Connection::OnClosingLogicalChannel(H323Channel & channel)
{
	WRAPTRACE(2, "Closing logical channel...");
//	if (channel.GetCodec() != NULL) {
//		(channel.GetCodec())->CloseRawDataChannel();
//	}    

	return H323Connection::OnClosingLogicalChannel(channel);
}

BOOL
WrapH323Connection::OnControlProtocolError(ControlProtocolErrors errorSource,
				const void *errorData)
{
	cout << "*** [" << GetCallToken() << "] H.323 CONTROL PROTOCOL ERROR ";
	switch (errorSource) {
		case H323Connection::e_MasterSlaveDetermination:
			cout << "(Master-Slave Determination)" << endl;
			break;
		case H323Connection::e_CapabilityExchange:
			cout << "(Capability Exchange)" << endl;
			break;
		case H323Connection::e_LogicalChannel:
			cout << "(Logical Channel)" << endl;
			break;
		case H323Connection::e_ModeRequest:
			cout << "(Mode Request)" << endl;
			break;
		case H323Connection::e_RoundTripDelay:
			cout << "(Roundtrip Delay)" << endl;
			break;
		default:
			cout << "(Unknown)" << endl;
			break;
	}

	// Special case for round trip delay errors
	if ((errorSource == H323Connection::e_RoundTripDelay) &&
		(endpoint.ShouldClearCallOnRoundTripFail() == FALSE)) 
		return TRUE;

	// Default case, Terminate call
	ClearCall();
	return FALSE;
}

BOOL
WrapH323Connection::RequestModeChange(const PString & newModes)
{
	WRAPTRACE(2, "RequestMode initiated");
	return H323Connection::RequestModeChange(newModes);
}

BOOL
WrapH323Connection::OnRequestModeChange(const H245_RequestMode & pdu,
		H245_RequestModeAck & ack, H245_RequestModeReject & reject,
		PINDEX & selectedMode)
{
	WRAPTRACE(2, "RequestMode received");
	if (H323Connection::OnRequestModeChange(pdu, ack, reject, selectedMode) != TRUE) {
		WRAPTRACE(2, "Invalid capability requested");
		return FALSE;
	}
	WRAPTRACE(2, "New capability accepted");
	return TRUE;
}

void
WrapH323Connection::OnModeChanged(const H245_ModeDescription & newMode)
{
	WRAPTRACE(2, "Changing capability");
	H323Connection::OnModeChanged(newMode);
}

void
WrapH323Connection::OnAcceptModeChange(const H245_RequestModeAck & pdu)
{
	WRAPTRACE(2, "New mode accepted from remote");
}

void
WrapH323Connection::OnRefusedModeChange(const H245_RequestModeReject * pdu)
{
	if (pdu == NULL) {
		WRAPTRACE(2, "Request timed out");
		return;
	}
	WRAPTRACE(2, "New mode rejected from remote");
}

#if 0
BOOL
WrapH323Connection::OnCreateLogicalChannel(const H323Capability & capability,
							H323Channel::Directions dir, unsigned & errorCode)
{
	int chanBandwidth;
	int res;

	if (H323Connection::OnCreateLogicalChannel(capability, dir, errorCode) == FALSE)
		return FALSE;

	// Do the bandwidth check
	res = ((WrapH323EndPoint &)GetEndPoint()).GetBandwidthAvailable();
	if (res == 0) {
		errorCode = H245_OpenLogicalChannelReject_cause::e_insufficientBandwidth;
		WRAPTRACE(5, "Bandwidth limit violation!");
		cout << "Bandwidth limit violation!" << endl;
		return FALSE;
	}
	chanBandwidth = 0;
	OpalMediaFormat *mf = new OpalMediaFormat(capability.GetFormatName());
	cout << "Format is " << capability.GetFormatName() << endl;
	if (mf != NULL) {
		chanBandwidth = mf->GetBandwidth();
		cout << "BW is " << chanBandwidth << endl;
		delete mf;
	} else {
		WRAPTRACE(5, "Channel with unknown format " << capability.GetFormatName());
		cout << "Channel with unknown format " << capability.GetFormatName() << endl;
	}
	res -= chanBandwidth;
	if (res <= 0) {
		errorCode = H245_OpenLogicalChannelReject_cause::e_insufficientBandwidth;
		WRAPTRACE(5, "Bandwidth limit violation!!");
		cout << "Bandwidth limit violation!!" << endl;
		return FALSE;
	}
	return TRUE;
}
#endif

#if 0
H323Channel *
WrapH323Connection::CreateRealTimeLogicalChannel(const H323Capability & capability,
							 H323Channel::Directions dir,
							 unsigned sessionID,
							 const H245_H2250LogicalChannelParameters * param)
{
	WRAPTRACE(3, "Creating external RTP channel, dir=" << dir << 
				", sessionid=" << sessionID);
	PIPSocket::Address * pip = new PIPSocket::Address("192.168.1.31");
	WORD port = 40404;
	return new H323_ExternalRTPChannel(*this, capability, dir, sessionID, *pip, port);
	//return new H323_ExternalRTPChannel(*this, capability, dir, sessionID);
}
#endif
// End of file //////////////////////////////////////////////////////////////
